@@ -42,6 +42,7 @@ static PyObject* UUID = NULL;
 static PyObject* Timestamp = NULL;
 static PyObject* MinKey = NULL;
 static PyObject* MaxKey = NULL;
+static PyObject* UTC = NULL;
 static PyTypeObject* REType = NULL;
 
 #if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
@@ -140,7 +141,8 @@ typedef struct {
 
 static int write_dict(bson_buffer* buffer, PyObject* dict,
                       unsigned char check_keys, unsigned char top_level);
-static PyObject* elements_to_dict(const char* string, int max, PyObject* as_class);
+static PyObject* elements_to_dict(const char* string, int max,
+                                  PyObject* as_class, unsigned char tz_aware);
 
 static bson_buffer* buffer_new(void) {
     bson_buffer* buffer;
@@ -276,6 +278,7 @@ static int _reload_python_objects(void) {
         _reload_object(&Timestamp, "pymongo.timestamp", "Timestamp") ||
         _reload_object(&MinKey, "pymongo.min_key", "MinKey") ||
         _reload_object(&MaxKey, "pymongo.max_key", "MaxKey") ||
+        _reload_object(&UTC, "pymongo.tz_util", "utc") ||
         _reload_object(&RECompile, "re", "compile")) {
         return 1;
     }
@@ -296,10 +299,13 @@ static int _reload_python_objects(void) {
  *
  * returns 0 on failure */
 static int write_element_to_buffer(bson_buffer* buffer, int type_byte, PyObject* value, unsigned char check_keys, unsigned char first_attempt) {
-    /* TODO this isn't quite the same as the Python version:
-     * here we check for type equivalence, not isinstance in some
-     * places. */
-    if (PyInt_CheckExact(value) || PyLong_CheckExact(value)) {
+    if (PyBool_Check(value)) {
+        const long bool = PyInt_AsLong(value);
+        const char c = bool ? 0x01 : 0x00;
+        *(buffer->buffer + type_byte) = 0x08;
+        return buffer_write_bytes(buffer, &c, 1);
+    }
+    else if (PyInt_Check(value) || PyLong_Check(value)) {
         const long long_value = PyInt_AsLong(value);
         const int int_value = (int)long_value;
         if (PyErr_Occurred() || long_value != int_value) { /* Overflow */
@@ -316,12 +322,7 @@ static int write_element_to_buffer(bson_buffer* buffer, int type_byte, PyObject*
         }
         *(buffer->buffer + type_byte) = 0x10;
         return buffer_write_bytes(buffer, (const char*)&int_value, 4);
-    } else if (PyBool_Check(value)) {
-        const long bool = PyInt_AsLong(value);
-        const char c = bool ? 0x01 : 0x00;
-        *(buffer->buffer + type_byte) = 0x08;
-        return buffer_write_bytes(buffer, &c, 1);
-    } else if (PyFloat_CheckExact(value)) {
+    } else if (PyFloat_Check(value)) {
         const double d = PyFloat_AsDouble(value);
         *(buffer->buffer + type_byte) = 0x01;
         return buffer_write_bytes(buffer, (const char*)&d, 8);
@@ -511,7 +512,7 @@ static int write_element_to_buffer(bson_buffer* buffer, int type_byte, PyObject*
         result = write_string(buffer, encoded);
         Py_DECREF(encoded);
         return result;
-    } else if (PyDateTime_CheckExact(value)) {
+    } else if (PyDateTime_Check(value)) {
         long long millis;
         PyObject* utcoffset = PyObject_CallMethod(value, "utcoffset", NULL);
         if (utcoffset != Py_None) {
@@ -950,24 +951,70 @@ static PyObject* _cbson_dict_to_bson(PyObject* self, PyObject* args) {
 
 /* add a lastError message on the end of the buffer.
  * returns 0 on failure */
-static int add_last_error(bson_buffer* buffer, int request_id) {
+static int add_last_error(bson_buffer* buffer, int request_id, PyObject* args) {
                                  /* message length: 62 */
-    if (!buffer_write_bytes(buffer, "\x3E\x00\x00\x00", 4) ||
-        !buffer_write_bytes(buffer, (const char*)&request_id, 4) ||
-        !buffer_write_bytes(buffer,
-                            "\x00\x00\x00\x00" /* responseTo */
-                            "\xd4\x07\x00\x00" /* opcode */
-                            "\x00\x00\x00\x00" /* options */
-                            "admin.$cmd\x00"   /* collection name */
-                            "\x00\x00\x00\x00" /* skip */
-                            "\xFF\xFF\xFF\xFF" /* limit (-1) */
-                            "\x17\x00\x00\x00" /* {getlasterror: 1} */
-                            "\x10getlasterror" /* ... */
-                            "\x00\x01\x00\x00" /* ... */
-                            "\x00\x00",        /* ... */
-                            54)) {
+    int message_start;
+    int document_start;
+    int message_length;
+    int document_length;
+    PyObject* key;
+    PyObject* value;
+    Py_ssize_t pos = 0;
+    PyObject* one;
+
+    message_start = buffer_save_bytes(buffer, 4);
+    if (message_start == -1) {
         return 0;
     }
+    if (!buffer_write_bytes(buffer, (const char*)&request_id, 4) ||
+        !buffer_write_bytes(buffer,
+                            "\x00\x00\x00\x00"  /* responseTo */
+                            "\xd4\x07\x00\x00"  /* opcode */
+                            "\x00\x00\x00\x00"  /* options */
+                            "admin.$cmd\x00"    /* collection name */
+                            "\x00\x00\x00\x00"  /* skip */
+                            "\xFF\xFF\xFF\xFF", /* limit (-1) */
+                            31)) {
+        return 0;
+    }
+
+    /* save space for length */
+    document_start = buffer_save_bytes(buffer, 4);
+    if (document_length == -1) {
+        return 0;
+    }
+
+    /* getlasterror: 1 */
+    one = PyLong_FromLong(1);
+    if (!write_pair(buffer, "getlasterror", 12, one, 0, 1)) {
+        Py_DECREF(one);
+        return 0;
+    }
+    Py_DECREF(one);
+
+    /* getlasterror options */
+    while (PyDict_Next(args, &pos, &key, &value)) {
+        if (!decode_and_write_pair(buffer, key, value, 0, 0)) {
+            return 0;
+        }
+    }
+
+    /* EOD */
+    if (!buffer_write_bytes(buffer, "\x00", 1)) {
+        return 0;
+    }
+
+    message_length = buffer->position - message_start;
+    document_length = buffer->position - document_start;
+    if (document_length > 4 * 1024 * 1024) {
+        PyObject* InvalidDocument = _error("InvalidDocument");
+        PyErr_SetString(InvalidDocument, "document too large - "
+                        "BSON documents are limited to 4 MB");
+        Py_DECREF(InvalidDocument);
+        return 0;
+    }
+    memcpy(buffer->buffer + message_start, &message_length, 4);
+    memcpy(buffer->buffer + document_start, &document_length, 4);
     return 1;
 }
 
@@ -981,15 +1028,16 @@ static PyObject* _cbson_insert_message(PyObject* self, PyObject* args) {
     int i;
     unsigned char check_keys;
     unsigned char safe;
+    PyObject* last_error_args;
     bson_buffer* buffer;
     int length_location;
     PyObject* result;
 
-    if (!PyArg_ParseTuple(args, "et#Obb",
+    if (!PyArg_ParseTuple(args, "et#ObbO",
                           "utf-8",
                           &collection_name,
                           &collection_name_length,
-                          &docs, &check_keys, &safe)) {
+                          &docs, &check_keys, &safe, &last_error_args)) {
         return NULL;
     }
 
@@ -1037,7 +1085,7 @@ static PyObject* _cbson_insert_message(PyObject* self, PyObject* args) {
     memcpy(buffer->buffer + length_location, &buffer->position, 4);
 
     if (safe) {
-        if (!add_last_error(buffer, request_id)) {
+        if (!add_last_error(buffer, request_id, last_error_args)) {
             buffer_free(buffer);
             return NULL;
         }
@@ -1060,16 +1108,18 @@ static PyObject* _cbson_update_message(PyObject* self, PyObject* args) {
     unsigned char multi;
     unsigned char upsert;
     unsigned char safe;
+    PyObject* last_error_args;
     int options;
     bson_buffer* buffer;
     int length_location;
     PyObject* result;
 
-    if (!PyArg_ParseTuple(args, "et#bbOOb",
+    if (!PyArg_ParseTuple(args, "et#bbOObO",
                           "utf-8",
                           &collection_name,
                           &collection_name_length,
-                          &upsert, &multi, &spec, &doc, &safe)) {
+                          &upsert, &multi, &spec, &doc, &safe,
+                          &last_error_args)) {
         return NULL;
     }
 
@@ -1111,7 +1161,7 @@ static PyObject* _cbson_update_message(PyObject* self, PyObject* args) {
     memcpy(buffer->buffer + length_location, &buffer->position, 4);
 
     if (safe) {
-        if (!add_last_error(buffer, request_id)) {
+        if (!add_last_error(buffer, request_id, last_error_args)) {
             buffer_free(buffer);
             return NULL;
         }
@@ -1240,7 +1290,7 @@ static PyObject* _cbson_get_more_message(PyObject* self, PyObject* args) {
 }
 
 static PyObject* get_value(const char* buffer, int* position, int type,
-                           PyObject* as_class) {
+                           PyObject* as_class, unsigned char tz_aware) {
     PyObject* value;
     switch (type) {
     case 1:
@@ -1271,20 +1321,41 @@ static PyObject* get_value(const char* buffer, int* position, int type,
         {
             int size;
             memcpy(&size, buffer + *position, 4);
-            value = elements_to_dict(buffer + *position + 4, size - 5, as_class);
+            value = elements_to_dict(buffer + *position + 4, size - 5, as_class, tz_aware);
             if (!value) {
                 return NULL;
             }
 
             /* Decoding for DBRefs */
             if (strcmp(buffer + *position + 5, "$ref") == 0) { /* DBRef */
-                PyObject* id = PyDict_GetItemString(value, "$id");
+                PyObject* dbref;
                 PyObject* collection = PyDict_GetItemString(value, "$ref");
+                PyObject* id = PyDict_GetItemString(value, "$id");
                 PyObject* database = PyDict_GetItemString(value, "$db");
 
-                /* This works even if there is no $db since database will be NULL and
-                   the call will be as if there were only two arguments specified. */
-                value = PyObject_CallFunctionObjArgs(DBRef, collection, id, database, NULL);
+                Py_INCREF(collection);
+                PyDict_DelItemString(value, "$ref");
+                Py_INCREF(id);
+                PyDict_DelItemString(value, "$id");
+
+                if (database == NULL) {
+                    database = Py_None;
+                    Py_INCREF(database);
+                } else {
+                    Py_INCREF(database);
+                    PyDict_DelItemString(value, "$db");
+                }
+
+                dbref = PyObject_CallFunctionObjArgs(DBRef, collection, id, database, value, NULL);
+                Py_DECREF(value);
+                value = dbref;
+
+                Py_DECREF(id);
+                Py_DECREF(collection);
+                Py_DECREF(database);
+                if (!value) {
+                    return NULL;
+                }
             }
 
             *position += size;
@@ -1309,7 +1380,7 @@ static PyObject* get_value(const char* buffer, int* position, int type,
                 int type = (int)buffer[(*position)++];
                 int key_size = strlen(buffer + *position);
                 *position += key_size + 1; /* just skip the key, they're in order. */
-                to_append = get_value(buffer, position, type, as_class);
+                to_append = get_value(buffer, position, type, as_class, tz_aware);
                 if (!to_append) {
                     return NULL;
                 }
@@ -1404,8 +1475,45 @@ static PyObject* get_value(const char* buffer, int* position, int type,
         }
     case 9:
         {
-            value = datetime_from_millis(*(long long*)(buffer + *position));
+            PyObject* replace;
+            PyObject* args;
+            PyObject* kwargs;
+            PyObject* naive = datetime_from_millis(*(long long*)(buffer + *position));
             *position += 8;
+            if (!tz_aware) { /* In the naive case, we're done here. */
+                value = naive;
+                break;
+            }
+
+            if (!naive) {
+                return NULL;
+            }
+            replace = PyObject_GetAttrString(naive, "replace");
+            Py_DECREF(naive);
+            if (!replace) {
+                return NULL;
+            }
+            args = PyTuple_New(0);
+            if (!args) {
+                Py_DECREF(replace);
+                return NULL;
+            }
+            kwargs = PyDict_New();
+            if (!kwargs) {
+                Py_DECREF(replace);
+                Py_DECREF(args);
+                return NULL;
+            }
+            if (PyDict_SetItemString(kwargs, "tzinfo", UTC) == -1) {
+                Py_DECREF(replace);
+                Py_DECREF(args);
+                Py_DECREF(kwargs);
+                return NULL;
+            }
+            value = PyObject_Call(replace, args, kwargs);
+            Py_DECREF(replace);
+            Py_DECREF(args);
+            Py_DECREF(kwargs);
             break;
         }
     case 11:
@@ -1482,7 +1590,8 @@ static PyObject* get_value(const char* buffer, int* position, int type,
             *position += code_length + 1;
 
             memcpy(&scope_size, buffer + *position, 4);
-            scope = elements_to_dict(buffer + *position + 4, scope_size - 5, (PyObject*)&PyDict_Type);
+            scope = elements_to_dict(buffer + *position + 4, scope_size - 5,
+                                     (PyObject*)&PyDict_Type, tz_aware);
             if (!scope) {
                 Py_DECREF(code);
                 return NULL;
@@ -1549,7 +1658,8 @@ static PyObject* get_value(const char* buffer, int* position, int type,
     return value;
 }
 
-static PyObject* elements_to_dict(const char* string, int max, PyObject* as_class) {
+static PyObject* elements_to_dict(const char* string, int max,
+                                  PyObject* as_class, unsigned char tz_aware) {
     int position = 0;
     PyObject* dict = PyObject_CallObject(as_class, NULL);
     if (!dict) {
@@ -1564,7 +1674,7 @@ static PyObject* elements_to_dict(const char* string, int max, PyObject* as_clas
             return NULL;
         }
         position += name_length + 1;
-        value = get_value(string, &position, type, as_class);
+        value = get_value(string, &position, type, as_class, tz_aware);
         if (!value) {
             return NULL;
         }
@@ -1582,11 +1692,12 @@ static PyObject* _cbson_bson_to_dict(PyObject* self, PyObject* args) {
     const char* string;
     PyObject* bson;
     PyObject* as_class;
+    unsigned char tz_aware;
     PyObject* dict;
     PyObject* remainder;
     PyObject* result;
 
-    if (!PyArg_ParseTuple(args, "OO", &bson, &as_class)) {
+    if (!PyArg_ParseTuple(args, "OOb", &bson, &as_class, &tz_aware)) {
         return NULL;
     }
 
@@ -1625,7 +1736,7 @@ static PyObject* _cbson_bson_to_dict(PyObject* self, PyObject* args) {
         return NULL;
     }
 
-    dict = elements_to_dict(string + 4, size - 5, as_class);
+    dict = elements_to_dict(string + 4, size - 5, as_class, tz_aware);
     if (!dict) {
         return NULL;
     }
@@ -1648,8 +1759,9 @@ static PyObject* _cbson_to_dicts(PyObject* self, PyObject* args) {
     PyObject* dict;
     PyObject* result;
     PyObject* as_class = (PyObject*)&PyDict_Type;
+    unsigned char tz_aware = 1;
 
-    if (!PyArg_ParseTuple(args, "O|O", &bson, &as_class)) {
+    if (!PyArg_ParseTuple(args, "O|Ob", &bson, &as_class, &tz_aware)) {
         return NULL;
     }
 
@@ -1692,7 +1804,7 @@ static PyObject* _cbson_to_dicts(PyObject* self, PyObject* args) {
             return NULL;
         }
 
-        dict = elements_to_dict(string + 4, size - 5, as_class);
+        dict = elements_to_dict(string + 4, size - 5, as_class, tz_aware);
         if (!dict) {
             return NULL;
         }
