@@ -18,7 +18,8 @@
 BINARY_SUBTYPE = 0
 """BSON binary subtype for binary data.
 
-This is becomming the default subtype and should be the most commonly used.
+This is becomming the default subtype and should be the most commonly
+used.
 
 .. versionadded:: 1.5
 """
@@ -32,7 +33,8 @@ FUNCTION_SUBTYPE = 1
 OLD_BINARY_SUBTYPE = 2
 """Old BSON binary subtype for binary data.
 
-This is still the default subtype, but that is changing to :data:`BINARY_SUBTYPE`.
+This is still the default subtype, but that is changing to
+:data:`BINARY_SUBTYPE`.
 
 .. versionadded:: 1.7
 """
