@@ -40,15 +40,10 @@ from pymongo.bson import BSON, is_valid, _to_dicts
 from pymongo.errors import InvalidDocument, InvalidStringData
 from pymongo.max_key import MaxKey
 from pymongo.min_key import MinKey
+from pymongo.tz_util import (FixedOffset,
+                             utc)
 import qcheck
 
-class SomeZone(datetime.tzinfo):
-    def utcoffset(self, dt):
-        return datetime.timedelta(minutes=555)
-    def dst(self, dt):
-        # a fixed-offset class:  doesn't account for DST
-        return datetime.timedelta(0)
-
 class TestBSON(unittest.TestCase):
 
     def setUp(self):
@@ -174,7 +169,9 @@ class TestBSON(unittest.TestCase):
         helper(SON([(u'test dst', datetime.datetime(1993, 4, 4, 2))]))
         helper({"big float": float(10000000000)})
         helper({"ref": DBRef("coll", 5)})
+        helper({"ref": DBRef("coll", 5, foo="bar", bar=4)})
         helper({"ref": DBRef("coll", 5, "foo")})
+        helper({"ref": DBRef("coll", 5, "foo", foo="bar")})
         helper({"ref": Timestamp(1,2)})
         helper({"foo": MinKey()})
         helper({"foo": MaxKey()})
@@ -186,12 +183,24 @@ class TestBSON(unittest.TestCase):
                               qcheck.gen_mongo_dict(3))
 
     def test_aware_datetime(self):
-        aware = datetime.datetime(1993, 4, 4, 2, tzinfo=SomeZone())
-        utc_naive = (aware - aware.utcoffset()).replace(tzinfo=None)
-        self.assertEqual(datetime.datetime(1993, 4, 3, 16, 45), utc_naive)
+        aware = datetime.datetime(1993, 4, 4, 2, tzinfo=FixedOffset(555, "SomeZone"))
+        as_utc = (aware - aware.utcoffset()).replace(tzinfo=utc)
+        self.assertEqual(datetime.datetime(1993, 4, 3, 16, 45, tzinfo=utc), as_utc)
+        after = BSON.from_dict({"date": aware}).to_dict(tz_aware=True)["date"]
+        self.assertEqual(utc, after.tzinfo)
+        self.assertEqual(as_utc, after)
+
+    def test_naive_decode(self):
+        aware = datetime.datetime(1993, 4, 4, 2, tzinfo=FixedOffset(555, "SomeZone"))
+        naive_utc = (aware - aware.utcoffset()).replace(tzinfo=None)
+        self.assertEqual(datetime.datetime(1993, 4, 3, 16, 45), naive_utc)
         after = BSON.from_dict({"date": aware}).to_dict()["date"]
         self.assertEqual(None, after.tzinfo)
-        self.assertEqual(utc_naive, after)
+        self.assertEqual(naive_utc, after)
+
+    def test_dst(self):
+        d = {"x": datetime.datetime(1993, 4, 4, 2)}
+        self.assertEqual(d, BSON.from_dict(d).to_dict())
 
     def test_bad_encode(self):
         self.assertRaises(InvalidStringData, BSON.from_dict,
@@ -289,6 +298,27 @@ class TestBSON(unittest.TestCase):
 
         self.assertEqual(1, BSON.from_dict({"x": 1}).to_dict(SON)["x"])
 
+        x = BSON.from_dict({"x": [{"y": 1}]})
+        self.assert_(isinstance(x.to_dict(SON)["x"][0], SON))
+
+    def test_subclasses(self):
+        # make sure we can serialize subclasses of native Python types.
+        class _myint(int):
+            pass
+        class _myfloat(float):
+            pass
+        class _myunicode(unicode):
+            pass
+        d = {'a' : _myint(42), 'b' : _myfloat(63.9),
+             'c' : _myunicode('hello world')
+            }
+        d2 = BSON.from_dict(d).to_dict()
+        for key, value in d2.iteritems():
+            orig_value = d[key]
+            orig_type = orig_value.__class__.__bases__[0]
+            self.assertEqual(type(value), orig_type)
+            self.assertEqual(value, orig_type(value))
+
 
 if __name__ == "__main__":
     unittest.main()
