@@ -35,7 +35,8 @@ from pymongo.errors import (DuplicateKeyError,
                             InvalidDocument,
                             InvalidName,
                             InvalidOperation,
-                            OperationFailure)
+                            OperationFailure,
+                            TimeoutError)
 from pymongo.objectid import ObjectId
 from pymongo.son import SON
 from test.test_connection import get_connection
@@ -513,6 +514,14 @@ class TestCollection(unittest.TestCase):
         self.assertRaises(expected_error,
                           db.test.update, {"x": 1}, {"$inc": {"x": 1}}, safe=True)
 
+    def test_error_code(self):
+        try:
+            self.db.test.update({}, {"$thismodifierdoesntexist": 1}, safe=True)
+            self.fail()
+        except OperationFailure, e:
+            if version.at_least(self.db.connection, (1, 3)):
+                self.assertEqual(10147, e.code)
+
 
     def test_index_on_subfield(self):
         db = self.db
@@ -641,6 +650,19 @@ class TestCollection(unittest.TestCase):
         self.assertEqual(2, db.test.remove({}, safe=True)["n"])
         self.assertEqual(0, db.test.remove({}, safe=True)["n"])
 
+    def test_last_error_options(self):
+        if not version.at_least(self.connection, (1, 5, 1)):
+            raise SkipTest()
+
+        self.assertRaises(TimeoutError, self.db.test.save, {"x": 1}, w=2, wtimeout=1)
+        self.assertRaises(TimeoutError, self.db.test.insert, {"x": 1}, w=2, wtimeout=1)
+        self.assertRaises(TimeoutError, self.db.test.update, {"x": 1}, {"y": 2}, w=2, wtimeout=1)
+        self.assertRaises(TimeoutError, self.db.test.remove, {"x": 1}, {"y": 2}, w=2, wtimeout=1)
+        self.db.test.save({"x": 1}, w=1, wtimeout=1)
+        self.db.test.insert({"x": 1}, w=1, wtimeout=1)
+        self.db.test.remove({"x": 1}, w=1, wtimeout=1)
+        self.db.test.update({"x": 1}, {"y": 2}, w=1, wtimeout=1)
+
     def test_count(self):
         db = self.db
         db.drop_collection("test")
