@@ -14,6 +14,7 @@
 
 """Test the connection module."""
 
+import datetime
 import os
 import sys
 import time
@@ -23,14 +24,17 @@ sys.path[0:0] = [""]
 
 from nose.plugins.skip import SkipTest
 
-from pymongo.connection import Connection
+from pymongo.connection import (Connection,
+                                _parse_uri)
 from pymongo.database import Database
 from pymongo.errors import (AutoReconnect,
+                            ConfigurationError,
                             ConnectionFailure,
                             InvalidName,
                             InvalidURI,
                             OperationFailure)
 from pymongo.son import SON
+from pymongo.tz_util import utc
 from test import version
 
 
@@ -49,11 +53,12 @@ class TestConnection(unittest.TestCase):
     def test_types(self):
         self.assertRaises(TypeError, Connection, 1)
         self.assertRaises(TypeError, Connection, 1.14)
-        self.assertRaises(TypeError, Connection, [])
         self.assertRaises(TypeError, Connection, "localhost", "27017")
         self.assertRaises(TypeError, Connection, "localhost", 1.14)
         self.assertRaises(TypeError, Connection, "localhost", [])
 
+        self.assertRaises(ConfigurationError, Connection, [])
+
     def test_constants(self):
         Connection.HOST = self.host
         Connection.PORT = self.port
@@ -75,6 +80,11 @@ class TestConnection(unittest.TestCase):
 
         self.assert_(Connection(self.host, self.port))
 
+    def test_host_w_port(self):
+        self.assert_(Connection("%s:%d" % (self.host, self.port)))
+        self.assertRaises(ConnectionFailure, Connection,
+                          "%s:1234567" % self.host, self.port)
+
     def test_repr(self):
         self.assertEqual(repr(Connection(self.host, self.port)),
                          "Connection('%s', %s)" % (self.host, self.port))
@@ -82,6 +92,7 @@ class TestConnection(unittest.TestCase):
     def test_getters(self):
         self.assertEqual(Connection(self.host, self.port).host, self.host)
         self.assertEqual(Connection(self.host, self.port).port, self.port)
+        self.assertEqual(set([(self.host, self.port)]), Connection(self.host, self.port).nodes)
 
     def test_get_db(self):
         connection = Connection(self.host, self.port)
@@ -229,38 +240,46 @@ class TestConnection(unittest.TestCase):
 
     def test_parse_uri(self):
         self.assertEqual(([("localhost", 27017)], None, None, None),
-                         Connection._parse_uri("localhost"))
-        self.assertRaises(InvalidURI, Connection._parse_uri, "http://foobar.com")
-        self.assertRaises(InvalidURI, Connection._parse_uri, "http://foo@foobar.com")
+                         _parse_uri("localhost", 27017))
+        self.assertEqual(([("localhost", 27018)], None, None, None),
+                         _parse_uri("localhost", 27018))
+        self.assertRaises(InvalidURI, _parse_uri,
+                          "http://foobar.com", 27017)
+        self.assertRaises(InvalidURI, _parse_uri,
+                          "http://foo@foobar.com", 27017)
 
         self.assertEqual(([("localhost", 27017)], None, None, None),
-                         Connection._parse_uri("mongodb://localhost"))
+                         _parse_uri("mongodb://localhost", 27017))
         self.assertEqual(([("localhost", 27017)], None, "fred", "foobar"),
-                         Connection._parse_uri("mongodb://fred:foobar@localhost"))
+                         _parse_uri("mongodb://fred:foobar@localhost",
+                                               27017))
         self.assertEqual(([("localhost", 27017)], "baz", "fred", "foobar"),
-                         Connection._parse_uri("mongodb://fred:foobar@localhost/baz"))
+                         _parse_uri("mongodb://fred:foobar@localhost/baz",
+                                               27017))
         self.assertEqual(([("example1.com", 27017), ("example2.com", 27017)],
                           None, None, None),
-                         Connection._parse_uri("mongodb://example1.com:27017,example2.com:27017"))
+                         _parse_uri("mongodb://example1.com:27017,example2.com:27017",
+                                               27018))
         self.assertEqual(([("localhost", 27017),
                            ("localhost", 27018),
                            ("localhost", 27019)], None, None, None),
-                         Connection._parse_uri("mongodb://localhost,localhost:27018,localhost:27019"))
+                         _parse_uri("mongodb://localhost,localhost:27018,localhost:27019",
+                                               27017))
 
         self.assertEqual(([("localhost", 27018)], None, None, None),
-                         Connection._parse_uri("localhost:27018"))
+                         _parse_uri("localhost:27018", 27017))
         self.assertEqual(([("localhost", 27017)], "foo", None, None),
-                         Connection._parse_uri("localhost/foo"))
+                         _parse_uri("localhost/foo", 27017))
         self.assertEqual(([("localhost", 27017)], None, None, None),
-                         Connection._parse_uri("localhost/"))
+                         _parse_uri("localhost/", 27017))
 
     def test_from_uri(self):
         c = Connection(self.host, self.port)
 
-        self.assertRaises(InvalidURI, Connection.from_uri, "mongodb://localhost/baz")
+        self.assertRaises(InvalidURI, Connection, "mongodb://localhost/baz")
 
-        self.assertEqual(c, Connection.from_uri("mongodb://%s:%s" %
-                                                (self.host, self.port)))
+        self.assertEqual(c, Connection("mongodb://%s:%s" %
+                                       (self.host, self.port)))
 
         c.admin.system.users.remove({})
         c.pymongo_test.system.users.remove({})
@@ -268,26 +287,26 @@ class TestConnection(unittest.TestCase):
         c.admin.add_user("admin", "pass")
         c.pymongo_test.add_user("user", "pass")
 
-        self.assertRaises(InvalidURI, Connection.from_uri,
+        self.assertRaises(ConfigurationError, Connection,
                           "mongodb://foo:bar@%s:%s" % (self.host, self.port))
-        self.assertRaises(InvalidURI, Connection.from_uri,
+        self.assertRaises(ConfigurationError, Connection,
                           "mongodb://admin:bar@%s:%s" % (self.host, self.port))
-        self.assertRaises(InvalidURI, Connection.from_uri,
+        self.assertRaises(ConfigurationError, Connection,
                           "mongodb://user:pass@%s:%s" % (self.host, self.port))
-        Connection.from_uri("mongodb://admin:pass@%s:%s" % (self.host, self.port))
+        Connection("mongodb://admin:pass@%s:%s" % (self.host, self.port))
 
-        self.assertRaises(InvalidURI, Connection.from_uri,
+        self.assertRaises(ConfigurationError, Connection,
                           "mongodb://admin:pass@%s:%s/pymongo_test" %
                           (self.host, self.port))
-        self.assertRaises(InvalidURI, Connection.from_uri,
+        self.assertRaises(ConfigurationError, Connection,
                           "mongodb://user:foo@%s:%s/pymongo_test" %
                           (self.host, self.port))
-        Connection.from_uri("mongodb://user:pass@%s:%s/pymongo_test" %
-                            (self.host, self.port))
+        Connection("mongodb://user:pass@%s:%s/pymongo_test" %
+                   (self.host, self.port))
 
-        self.assert_(Connection.from_uri("mongodb://%s:%s" %
-                                         (self.host, self.port),
-                                         slave_okay=True).slave_okay)
+        self.assert_(Connection("mongodb://%s:%s" %
+                                (self.host, self.port),
+                                slave_okay=True).slave_okay)
 
     def test_fork(self):
         """Test using a connection before and after a fork.
@@ -376,32 +395,45 @@ class TestConnection(unittest.TestCase):
         self.assert_(isinstance(db.test.find_one(), dict))
         self.failIf(isinstance(db.test.find_one(), SON))
 
-# TODO come up with a different way to test `network_timeout`. This is just
-# too sketchy.
-#
-#     def test_socket_timeout(self):
-#         no_timeout = Connection(self.host, self.port)
-#         timeout = Connection(self.host, self.port, network_timeout=0.1)
+    def test_network_timeout(self):
+        no_timeout = Connection(self.host, self.port)
+        timeout = Connection(self.host, self.port, network_timeout=0.1)
 
-#         no_timeout.pymongo_test.drop_collection("test")
+        no_timeout.pymongo_test.drop_collection("test")
+        no_timeout.pymongo_test.test.insert({"x": 1}, safe=True)
 
-#         no_timeout.pymongo_test.test.save({"x": 1})
+        where_func = """function (doc) {
+  var d = new Date().getTime() + 200;
+  var x = new Date().getTime();
+  while (x < d) {
+    x = new Date().getTime();
+  }
+  return true;
+}"""
 
-#         where_func = """function (doc) {
-#   var d = new Date().getTime() + 1000;
-#   var x = new Date().getTime();
-#   while (x < d) {
-#     x = new Date().getTime();
-#   }
-#   return true;
-# }"""
+        def get_x(db):
+            return db.test.find().where(where_func).next()["x"]
+        self.assertEqual(1, get_x(no_timeout.pymongo_test))
+        self.assertRaises(ConnectionFailure, get_x, timeout.pymongo_test)
 
-#         def get_x(db):
-#             return db.test.find().where(where_func).next()["x"]
+        def get_x_timeout(db, t):
+            return db.test.find(network_timeout=t).where(where_func).next()["x"]
+        self.assertEqual(1, get_x_timeout(timeout.pymongo_test, None))
+        self.assertRaises(ConnectionFailure, get_x_timeout,
+                          no_timeout.pymongo_test, 0.1)
 
-#         self.assertEqual(1, get_x(no_timeout.pymongo_test))
-#         self.assertRaises(ConnectionFailure, get_x, timeout.pymongo_test)
-#         self.assertEqual(1, no_timeout.pymongo_test.test.find().next()["x"])
+    def test_tz_aware(self):
+        aware = Connection(self.host, self.port, tz_aware=True)
+        naive = Connection(self.host, self.port)
+        aware.pymongo_test.drop_collection("test")
+
+        now = datetime.datetime.utcnow()
+        aware.pymongo_test.test.insert({"x": now}, safe=True)
+
+        self.assertEqual(None, naive.pymongo_test.test.find_one()["x"].tzinfo)
+        self.assertEqual(utc, aware.pymongo_test.test.find_one()["x"].tzinfo)
+        self.assertEqual(aware.pymongo_test.test.find_one()["x"].replace(tzinfo=None),
+                         naive.pymongo_test.test.find_one()["x"])
 
 
 if __name__ == "__main__":
